@@ -17,6 +17,30 @@
  *
  * Change History:
  * $Log: mbox.c,v $
+ * Revision 1.83  2004/06/28 11:44:45  nigelhorne
+ * Remove empty parts
+ *
+ * Revision 1.82  2004/06/25 13:56:38  nigelhorne
+ * Optimise messages without other messages encapsulated within them
+ *
+ * Revision 1.81  2004/06/24 21:36:38  nigelhorne
+ * Plug memory leak with large number of attachments
+ *
+ * Revision 1.80  2004/06/23 16:23:25  nigelhorne
+ * Further empty line optimisation
+ *
+ * Revision 1.79  2004/06/22 04:08:01  nigelhorne
+ * Optimise empty lines
+ *
+ * Revision 1.78  2004/06/21 10:21:19  nigelhorne
+ * Fix crash when a multipart/mixed message contains many parts that need to be scanned as attachments
+ *
+ * Revision 1.77  2004/06/18 10:07:12  nigelhorne
+ * Allow any number of alternatives in multipart messages
+ *
+ * Revision 1.76  2004/06/16 08:07:39  nigelhorne
+ * Added thread safety
+ *
  * Revision 1.75  2004/06/14 09:07:10  nigelhorne
  * Handle spam using broken e-mail generators for multipart/alternative
  *
@@ -213,7 +237,7 @@
  * Compilable under SCO; removed duplicate code with message.c
  *
  */
-static	char	const	rcsid[] = "$Id: mbox.c,v 1.75 2004/06/14 09:07:10 nigelhorne Exp $";
+static	char	const	rcsid[] = "$Id: mbox.c,v 1.83 2004/06/28 11:44:45 nigelhorne Exp $";
 
 #if HAVE_CONFIG_H
 #include "clamav-config.h"
@@ -244,6 +268,10 @@ static	char	const	rcsid[] = "$Id: mbox.c,v 1.75 2004/06/14 09:07:10 nigelhorne E
 #include <sys/param.h>
 #include <clamav.h>
 
+#ifdef	CL_THREAD_SAFE
+#include <pthread.h>
+#endif
+
 #include "table.h"
 #include "mbox.h"
 #include "blob.h"
@@ -253,6 +281,21 @@ static	char	const	rcsid[] = "$Id: mbox.c,v 1.75 2004/06/14 09:07:10 nigelhorne E
 #include "defaults.h"
 #include "str.h"
 
+#ifdef	CL_DEBUG
+#if __GLIBC__ == 2 && __GLIBC_MINOR__ >= 1
+#define HAVE_BACKTRACE
+#endif
+
+#ifdef HAVE_BACKTRACE
+#include <execinfo.h>
+#include <signal.h>
+#include <syslog.h>
+#endif
+
+static	void	sigsegv(int sig);
+static	void	print_trace(int use_syslog);
+#endif
+
 #if	defined(NO_STRTOK_R) || !defined(CL_THREAD_SAFE)
 #undef strtok_r
 #undef __strtok_r
@@ -269,7 +312,7 @@ static	char	const	rcsid[] = "$Id: mbox.c,v 1.75 2004/06/14 09:07:10 nigelhorne E
 
 typedef enum    { FALSE = 0, TRUE = 1 } bool;
 
-static	message	*parseEmailHeaders(const message *m, const table_t *rfc821Table);
+static	message	*parseEmailHeaders(message *m, const table_t *rfc821Table, bool destroy);
 static	int	parseEmailHeader(message *m, const char *line, const table_t *rfc821Table);
 static	int	parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, const char *dir, table_t *rfc821Table, table_t *subtypeTable);
 static	int	boundaryStart(const char *line, const char *boundary);
@@ -333,6 +376,10 @@ static	const	struct tableinit {
 	{	"appledouble",	APPLEDOUBLE	},
 	{	NULL,		0		}
 };
+
+#ifdef	CL_THREAD_SAFE
+static	pthread_mutex_t	tables_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
 static	table_t	*rfc821Table, *subtypeTable;
 
 /* Maximum filenames under various systems */
@@ -378,6 +425,9 @@ cl_mbox(const char *dir, int desc)
 	message *m, *body;
 	FILE *fd;
 	char buffer[LINE_LENGTH];
+#ifdef	CL_DEBUG
+	void (*segv)(int);
+#endif
 
 	cli_dbgmsg("in mbox()\n");
 
@@ -398,15 +448,30 @@ cl_mbox(const char *dir, int desc)
 		return 0;
 	}
 
+#ifdef	CL_THREAD_SAFE
+	pthread_mutex_lock(&tables_mutex);
+#endif
 	if(rfc821Table == NULL) {
 		assert(subtypeTable == NULL);
 
 		if(initialiseTables(&rfc821Table, &subtypeTable) < 0) {
+			rfc821Table = NULL;
+			subtypeTable = NULL;
+#ifdef	CL_THREAD_SAFE
+			pthread_mutex_unlock(&tables_mutex);
+#endif
 			messageDestroy(m);
 			fclose(fd);
 			return -1;
 		}
 	}
+#ifdef	CL_THREAD_SAFE
+	pthread_mutex_unlock(&tables_mutex);
+#endif
+
+#ifdef	CL_DEBUG
+	segv = signal(SIGSEGV, sigsegv);
+#endif
 
 	/*
 	 * is it a UNIX style mbox with more than one
@@ -429,7 +494,7 @@ cl_mbox(const char *dir, int desc)
 				/*
 				 * End of a message in the mail box
 				 */
-				body = parseEmailHeaders(m, rfc821Table);
+				body = parseEmailHeaders(m, rfc821Table, TRUE);
 				messageDestroy(m);
 				if(messageGetBody(body))
 					if(!parseEmailBody(body,  NULL, 0, NULL, dir, rfc821Table, subtypeTable)) {
@@ -475,7 +540,7 @@ cl_mbox(const char *dir, int desc)
 
 	retcode = 0;
 
-	body = parseEmailHeaders(m, rfc821Table);
+	body = parseEmailHeaders(m, rfc821Table, TRUE);
 	messageDestroy(m);
 	/*
 	 * Write out the last entry in the mailbox
@@ -491,6 +556,10 @@ cl_mbox(const char *dir, int desc)
 
 	cli_dbgmsg("cli_mbox returning %d\n", retcode);
 
+#ifdef	CL_DEBUG
+	signal(SIGSEGV, segv);
+#endif
+
 	return retcode;
 }
 
@@ -500,63 +569,85 @@ cl_mbox(const char *dir, int desc)
  * This function parses the headers of m and sets the message's arguments
  *
  * Returns the message's body with the correct arguments set
+ *
+ * The downside of this approach is that for a short time we have two copies
+ * of the message in memory, the upside is that it makes for easier parsing
+ * of encapsulated messages, and in the long run uses less memory in those
+ * scenarios
+ * BUT: if 'destroy' is set, the caller has given us a hint than 'm' will
+ * not be used again before it is destroyed, so we can trash it
  */
 static message *
-parseEmailHeaders(const message *m, const table_t *rfc821Table)
+parseEmailHeaders(message *m, const table_t *rfc821Table, bool destroy)
 {
 	bool inContinuationHeader = FALSE;	/* state machine: ugh */
 	bool inHeader = TRUE;
-	const text *t;
+	text *t;
 	message *ret;
 
+	cli_dbgmsg("parseEmailHeaders\n");
+
 	if(m == NULL)
 		return NULL;
 
 	ret = messageCreate();
 
-	for(t = messageGetBody(m); t; t = t->t_next) {
-		char *buffer = strdup(t->t_text);
+	for(t = (text *)messageGetBody(m); t; t = t->t_next) {
+		char *buffer;
 #ifdef CL_THREAD_SAFE
 		char *strptr;
 #endif
 
-		if(buffer == NULL)
-			break;
-
-		cli_chomp(buffer);
+		if(t->t_text) {
+			if(destroy) {
+				buffer = t->t_text;
+				t->t_text = NULL;
+			} else {
+				buffer = strdup(t->t_text);
+				if(buffer == NULL)
+					break;
+			}
+			if(cli_chomp(buffer) == 0) {
+				free(buffer);
+				buffer = NULL;
+			}
+		} else
+			buffer = NULL;
 
 		/*
 		 * Section B.2 of RFC822 says TAB or SPACE means
 		 * a continuation of the previous entry.
 		 */
-		if(inHeader && ((buffer[0] == '\t') || (buffer[0] == ' ')))
+		if(inHeader && buffer &&
+		  ((buffer[0] == '\t') || (buffer[0] == ' ')))
 			inContinuationHeader = TRUE;
 
 		if(inContinuationHeader) {
-			const char *ptr;
-
 			if(!continuationMarker(buffer))
 				inContinuationHeader = FALSE;	 /* no more args */
 
 			/*
 			 * Add all the arguments on the line
 			 */
-			for(ptr = strtok_r(buffer, ";", &strptr); ptr; ptr = strtok_r(NULL, ":", &strptr))
-				messageAddArgument(ret, ptr);
-			free(buffer);
-		} else if(inHeader) {
-			cli_dbgmsg("Deal with header %s\n", buffer);
+			if(buffer) {
+				const char *ptr;
 
+				for(ptr = strtok_r(buffer, ";", &strptr); ptr; ptr = strtok_r(NULL, ":", &strptr))
+					messageAddArgument(ret, ptr);
+				free(buffer);
+			}
+		} else if(inHeader) {
 			/*
 			 * A blank line signifies the end of the header and
 			 * the start of the text
 			 */
-			if(strlen(buffer) == 0) {
+			if(buffer == NULL) {
 				cli_dbgmsg("End of header information\n");
 				inContinuationHeader = inHeader = FALSE;
 			} else if(parseEmailHeader(ret, buffer, rfc821Table) == CONTENT_TYPE)
 				inContinuationHeader = continuationMarker(buffer);
-			free(buffer);
+			if(buffer)
+				free(buffer);
 		} else {
 			/*cli_dbgmsg("Add line to body '%s'\n", buffer);*/
 			messageAddLine(ret, buffer, 0);
@@ -589,6 +680,9 @@ parseEmailHeader(message *m, const char *line, const table_t *rfc821Table)
 
 	copy = strdup(line);
 
+	if(copy == NULL)
+		return -1;
+
 	cmd = strtok_r(copy, ":", &strptr);
 
 	if(cmd && *cmd) {
@@ -629,7 +723,7 @@ parseEmailHeader(message *m, const char *line, const table_t *rfc821Table)
 static int	/* success or fail */
 parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, const char *dir, table_t *rfc821Table, table_t *subtypeTable)
 {
-	message *messages[MAXALTERNATIVE];
+	message **messages;	/* parts of a multipart message */
 	int inhead, inMimeHead, i, rc = 1, htmltextPart, multiparts = 0;
 	text *aText;
 	blob *blobList[MAX_ATTACHMENTS], **blobs;
@@ -641,11 +735,12 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 	/* Pre-assertions */
 	if(nBlobs >= MAX_ATTACHMENTS) {
 		cli_warnmsg("Not all attachments will be scanned\n");
-		return 2;
+		/*return 2;*/
 	}
 
 	aText = textIn;
 	blobs = blobsIn;
+	messages = NULL;
 	mainMessage = messageIn;
 
 	/* Anything left to be parsed? */
@@ -745,9 +840,11 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 			 * This looks like parseEmailHeaders() - maybe there's
 			 * some duplication of code to be cleaned up
 			 */
-			for(multiparts = 0; t_line && (multiparts < MAXALTERNATIVE); multiparts++) {
+			for(multiparts = 0; t_line; multiparts++) {
 				int lines = 0;
 
+				messages = cli_realloc(messages, ((multiparts + 1) * sizeof(message *)));
+
 				aMessage = messages[multiparts] = messageCreate();
 
 				cli_dbgmsg("Now read in part %d\n", multiparts);
@@ -756,14 +853,15 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 				 * Ignore blank lines. There shouldn't be ANY
 				 * but some viruses insert them
 				 */
-				while((t_line = t_line->t_next) != NULL) {
-					cli_chomp(t_line->t_text);
-					if(strlen(t_line->t_text) != 0)
+				while((t_line = t_line->t_next) != NULL)
+					if(t_line->t_text &&
+					   (cli_chomp(t_line->t_text) > 0))
 						break;
-				}
 
 				if(t_line == NULL) {
 					cli_dbgmsg("Empty part\n");
+					messageDestroy(aMessage);
+					--multiparts;
 					continue;
 				}
 
@@ -774,6 +872,10 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 						inMimeHead, inhead, boundary, line, t_line->t_next ? t_line->t_next->t_text : "(null)");*/
 
 					if(inMimeHead) {
+						if(line == NULL) {
+							inhead = inMimeHead = 0;
+							continue;
+						}
 						/*
 						 * Handle continuation lines
 						 * because the previous line
@@ -805,7 +907,8 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 						inMimeHead = continuationMarker(line);
 						messageAddArgument(aMessage, line);
 					} else if(inhead) {
-						if(strlen(line) == 0) {
+						if(line == NULL) {
+							/* empty line */
 							inhead = 0;
 							continue;
 						}
@@ -841,7 +944,9 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 						 */
 						inMimeHead = continuationMarker(line);
 						if(!inMimeHead)
-							if(t_line->t_next && ((t_line->t_next->t_text[0] == '\t') || (t_line->t_next->t_text[0] == ' ')))
+							if(t_line->t_next &&
+							   t_line->t_next->t_text &&
+							  ((t_line->t_next->t_text[0] == '\t') || (t_line->t_next->t_text[0] == ' ')))
 								inMimeHead = TRUE;
 
 						parseEmailHeader(aMessage, line, rfc821Table);
@@ -881,8 +986,11 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 				mainMessage = NULL;
 			}
 
-			if(multiparts == 0)
+			if(multiparts == 0) {
+				if(messages)
+					free(messages);
 				return 2;	/* Nothing to do */
+			}
 
 			cli_dbgmsg("The message has %d parts\n", multiparts);
 			cli_dbgmsg("Find out the multipart type(%s)\n", mimeSubtype);
@@ -1136,7 +1244,7 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 						 * many nested levels are
 						 * involved.
 						 */
-						body = parseEmailHeaders(aMessage, rfc821Table);
+						body = parseEmailHeaders(aMessage, rfc821Table, TRUE);
 						/*
 						 * We've fininished with the
 						 * original copy of the message,
@@ -1162,7 +1270,7 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 						 */
 						cli_dbgmsg("Found multipart inside multipart\n");
 						if(aMessage) {
-							body = parseEmailHeaders(aMessage, rfc821Table);
+							body = parseEmailHeaders(aMessage, rfc821Table, TRUE);
 							if(body) {
 								assert(aMessage == messages[i]);
 								messageDestroy(messages[i]);
@@ -1211,9 +1319,16 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 					assert(addToText || addAttachment);
 					assert(!(addToText && addAttachment));
 
-					if(addToText) {
+					if(addToText)
 						aText = textAdd(aText, messageGetBody(aMessage));
-					} else if(addAttachment) {
+					else if(numberOfAttachments >= MAX_ATTACHMENTS) {
+						cli_warnmsg("Not all attachments will be scanned\n");
+						/*
+						 * Try our best to save it
+						 * somewhere
+						 */
+						aText = textAdd(aText, messageGetBody(aMessage));
+					} else {
 						blob *aBlob = messageToBlob(aMessage);
 
 						if(aBlob) {
@@ -1258,6 +1373,7 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 							blobGetFilename(blobs[i]),
 							blobGetFilename(blobList[j]));
 						blobDestroy(blobs[i]);
+						blobs[i] = NULL;
 					}
 				}
 
@@ -1336,6 +1452,9 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 			if(aText && (textIn == NULL))
 				textDestroy(aText);
 
+			if(messages)
+				free(messages);
+
 			return rc;
 
 		case MESSAGE:
@@ -1353,7 +1472,7 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 			}
 			if((strcasecmp(mimeSubtype, "rfc822") == 0) ||
 			   (strcasecmp(mimeSubtype, "delivery-status") == 0)) {
-				message *m = parseEmailHeaders(mainMessage, rfc821Table);
+				message *m = parseEmailHeaders(mainMessage, rfc821Table, FALSE);
 				if(m) {
 					cli_dbgmsg("Decode rfc822");
 
@@ -1369,7 +1488,7 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 				break;
 			} else if(strcasecmp(mimeSubtype, "partial") == 0)
 				/* TODO */
-				cli_warnmsg("Content-type message/partial not yet supported");
+				cli_warnmsg("Content-type message/partial not yet supported\n");
 			else if(strcasecmp(mimeSubtype, "external-body") == 0)
 				/*
 				 * I don't believe that we should be going
@@ -1382,6 +1501,8 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 
 			if(mainMessage && (mainMessage != messageIn))
 				messageDestroy(mainMessage);
+			if(messages)
+				free(messages);
 			return 0;
 
 		case APPLICATION:
@@ -1494,7 +1615,8 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 				 * won't be scanned
 				 */
 				for(t = t_line; t; t = t->t_next)
-					if((strncasecmp(t->t_text, encoding, sizeof(encoding) - 1) == 0) &&
+					if(t->t_text &&
+					   (strncasecmp(t->t_text, encoding, sizeof(encoding) - 1) == 0) &&
 					   (strstr(t->t_text, "7bit") == NULL))
 						break;
 				if(t && ((b = textToBlob(t_line, NULL)) != NULL)) {
@@ -1576,6 +1698,9 @@ parseEmailBody(message *messageIn, blob **blobsIn, int nBlobs, text *textIn, con
 	if(mainMessage && (mainMessage != messageIn))
 		messageDestroy(mainMessage);
 
+	if(messages)
+		free(messages);
+
 	cli_dbgmsg("parseEmailBody() returning %d\n", rc);
 
 	return rc;
@@ -1597,6 +1722,8 @@ boundaryStart(const char *line, const char *boundary)
 	 * notice the extra '-'
 	 */
 	/*cli_dbgmsg("boundaryStart: line = '%s' boundary = '%s'\n", line, boundary);*/
+	if(line == NULL)
+		return 0;	/* empty line */
 	if(strstr(line, boundary) != NULL) {
 		cli_dbgmsg("found %s in %s\n", boundary, line);
 		return 1;
@@ -1618,6 +1745,8 @@ endOfMessage(const char *line, const char *boundary)
 {
 	size_t len;
 
+	if(line == NULL)
+		return 0;
 	if(*line++ != '-')
 		return 0;
 	if(*line++ != '-')
@@ -1740,6 +1869,7 @@ strstrip(char *s)
 {
 	if(s == (char *)NULL)
 		return(0);
+
 	return(strip(s, strlen(s) + 1));
 }
 
@@ -1755,7 +1885,8 @@ continuationMarker(const char *line)
 {
 	const char *ptr;
 
-	assert(line != NULL);
+	if(line == NULL)
+		return FALSE;
 
 #ifdef	CL_DEBUG
 	cli_dbgmsg("continuationMarker(%s)\n", line);
@@ -1981,3 +2112,39 @@ saveFile(const blob *b, const char *dir)
 
 	return (close(fd) >= 0);
 }
+
+#ifdef HAVE_BACKTRACE
+static void
+sigsegv(int sig)
+{
+	signal(SIGSEGV, SIG_DFL);
+	print_trace(1);
+	exit(SIGSEGV);
+}
+
+static void
+print_trace(int use_syslog)
+{
+	void *array[10];
+	size_t size;
+	char **strings;
+	size_t i;
+	pid_t pid = getpid();
+
+	size = backtrace(array, 10);
+	strings = backtrace_symbols(array, size);
+
+	if(use_syslog == 0)
+		cli_dbgmsg("Backtrace of pid %d:\n", pid);
+	else
+		syslog(LOG_ERR, "Backtrace of pid %d:", pid);
+
+	for(i = 0; i < size; i++)
+		if(use_syslog)
+			syslog(LOG_ERR, "bt[%d]: %s", i, strings[i]);
+		else
+			cli_dbgmsg("%s\n", strings[i]);
+
+	free(strings);
+}
+#endif
