@@ -17,6 +17,18 @@
  *
  * Change History:
  * $Log: message.c,v $
+ * Revision 1.63  2004/06/26 13:16:25  nigelhorne
+ * Added newline to end of warning message
+ *
+ * Revision 1.62  2004/06/24 21:37:26  nigelhorne
+ * Handle uuencoded files created with buggy software
+ *
+ * Revision 1.61  2004/06/22 04:08:02  nigelhorne
+ * Optimise empty lines
+ *
+ * Revision 1.60  2004/06/16 08:07:39  nigelhorne
+ * Added thread safety
+ *
  * Revision 1.59  2004/06/02 10:11:09  nigelhorne
  * Corrupted binHex could crash on non Linux systems
  *
@@ -174,7 +186,7 @@
  * uuencodebegin() no longer static
  *
  */
-static	char	const	rcsid[] = "$Id: message.c,v 1.59 2004/06/02 10:11:09 nigelhorne Exp $";
+static	char	const	rcsid[] = "$Id: message.c,v 1.63 2004/06/26 13:16:25 nigelhorne Exp $";
 
 #if HAVE_CONFIG_H
 #include "clamav-config.h"
@@ -200,6 +212,10 @@ static	char	const	rcsid[] = "$Id: message.c,v 1.59 2004/06/02 10:11:09 nigelhorn
 #include <ctype.h>
 #include <stdio.h>
 
+#ifdef	CL_THREAD_SAFE
+#include <pthread.h>
+#endif
+
 #include "mbox.h"
 #include "table.h"
 #include "blob.h"
@@ -237,7 +253,7 @@ static	const	struct	encoding_map {
 	{	"7bit",			NOENCODING	},
 	{	"text/plain",		NOENCODING	},
 	{	"quoted-printable",	QUOTEDPRINTABLE	},	/* rfc1522 */
-	{	"base64",		BASE64		},
+	{	"base64",		BASE64		},	/* rfc2045 */
 	{	"8bit",			EIGHTBIT	},
 	{	"8 bit",		EIGHTBIT	},	/* incorrect */
 	{	"x-uuencode",		UUENCODE	},
@@ -310,6 +326,9 @@ messageReset(message *m)
 void
 messageSetMimeType(message *mess, const char *type)
 {
+#ifdef	CL_THREAD_SAFE
+	static pthread_mutex_t mime_mutex = PTHREAD_MUTEX_INITIALIZER;
+#endif
 	static table_t *mime_table;
 	int typeval;
 
@@ -325,19 +344,33 @@ messageSetMimeType(message *mess, const char *type)
 		if(*type++ == '\0')
 			return;
 
+#ifdef	CL_THREAD_SAFE
+	pthread_mutex_lock(&mime_mutex);
+#endif
 	if(mime_table == NULL) {
 		const struct mime_map *m;
 
 		mime_table = tableCreate();
-		if(mime_table == NULL)
+		if(mime_table == NULL) {
+#ifdef	CL_THREAD_SAFE
+			pthread_mutex_unlock(&mime_mutex);
+#endif
 			return;
+		}
 
 		for(m = mime_map; m->string; m++)
 			if(!tableInsert(mime_table, m->string, m->type)) {
 				tableDestroy(mime_table);
+				mime_table = NULL;
+#ifdef	CL_THREAD_SAFE
+				pthread_mutex_unlock(&mime_mutex);
+#endif
 				return;
 			}
 	}
+#ifdef	CL_THREAD_SAFE
+	pthread_mutex_unlock(&mime_mutex);
+#endif
 
 	typeval = tableFind(mime_table, type);
 
@@ -768,38 +801,26 @@ messageAddLine(message *m, const char *line, int takeCopy)
 
 	m->body_last->t_next = NULL;
 
-	if(takeCopy) {
-		m->body_last->t_text = strdup((line) ? line : "");
-		if(m->body_last->t_text == NULL) {
-			cli_errmsg("messageAddLine: out of memory\n");
-			return -1;
-		}
-		assert(m->body_last->t_text != NULL);
-	} else {
-		assert(line != NULL);
-		m->body_last->t_text = (char *)line;
-	}
+	if(line && *line) {
+		if(takeCopy) {
+			m->body_last->t_text = strdup(line);
+			if(m->body_last->t_text == NULL) {
+				cli_errmsg("messageAddLine: out of memory\n");
+				return -1;
+			}
+			// cli_chomp(m->body_last->t_text);
+		} else
+			m->body_last->t_text = (char *)line;
 
-	/*
-	 * See if this line marks the start of a non MIME inclusion that
-	 * will need to be scanned
-	 *
-	 * Notes that X- lines are not taken as start of mails because
-	 * cli_filetype() is too keen: any line it finds that starts X-
-	 * is seen to be the start of a new message, which results in too
-	 * many false positives for locating the possible start of a bounce,
-	 * and allows some instances of Worm.SomeFool.Gen-1 to get through in
-	 * bounce messages which end up not being correctly handled because
-	 * the real start of a bounce header is missed because of the earlier
-	 * false positive
-	 */
-	if(line) {
+		/*
+		 * See if this line marks the start of a non MIME inclusion that
+		 * will need to be scanned
+		 */
 		if((m->encoding == NULL) &&
 		   (strncasecmp(line, encoding, sizeof(encoding) - 1) == 0) &&
 		   (strstr(line, "7bit") == NULL))
 			m->encoding = m->body_last;
 		else if((m->bounce == NULL) &&
-			(strncmp(line, "X-", 2) != 0) &&	/*!!*/
 			(cli_filetype(line, strlen(line)) == CL_MAILFILE))
 				m->bounce = m->body_last;
 		else if((m->binhex == NULL) &&
@@ -812,7 +833,9 @@ messageAddLine(message *m, const char *line, int takeCopy)
 			(isdigit(line[8])) &&
 			(line[9] == ' ')))
 				m->uuencode = m->body_last;
-	}
+	} else
+		m->body_last->t_text = NULL;
+
 	return 1;
 }
 
@@ -947,7 +970,8 @@ messageToBlob(message *m)
 		 * See RFC1741
 		 */
 		while((t_line = t_line->t_next) != NULL)
-			blobAddData(tmp, (unsigned char *)t_line->t_text, strlen(t_line->t_text));
+			if(t_line->t_text)
+				blobAddData(tmp, (unsigned char *)t_line->t_text, strlen(t_line->t_text));
 
 		data = blobGetData(tmp);
 
@@ -1212,9 +1236,15 @@ messageToBlob(message *m)
 		unsigned char *uptr;
 		const char *line = t_line->t_text;
 
-		if(messageGetEncoding(m) == UUENCODE)
+		if(messageGetEncoding(m) == UUENCODE) {
+			/*
+			 * There should be no blank lines in uuencoded files...
+			 */
+			if(line == NULL)
+				continue;
 			if(strcasecmp(line, "end") == 0)
 				break;
+		}
 
 		uptr = decodeLine(m, line, data, sizeof(data));
 
@@ -1310,13 +1340,12 @@ messageToText(const message *m)
 				last = last->t_next;
 			}
 
-			if(last)
-				last->t_text = strdup((char *)data);
-
-			if((last == NULL) || (last->t_text == NULL))
+			if(last == NULL)
 				break;
 
-			if(messageGetEncoding(m) == BASE64)
+			last->t_text = data[0] ? strdup((char *)data) : NULL;
+
+			if(line && messageGetEncoding(m) == BASE64)
 				if(strchr(line, '='))
 					break;
 		}
@@ -1457,7 +1486,6 @@ decodeLine(const message *m, const char *line, unsigned char *buf, size_t buflen
 	char *copy;
 
 	assert(m != NULL);
-	assert(line != NULL);
 	assert(buf != NULL);
 
 	switch(messageGetEncoding(m)) {
@@ -1469,11 +1497,20 @@ decodeLine(const message *m, const char *line, unsigned char *buf, size_t buflen
 		case NOENCODING:
 		case EIGHTBIT:
 		default:	/* unknown encoding type - try our best */
+			if(line == NULL) {	/* empty line */
+				*buf++ = '\n';
+				break;
+			}
 			buf = (unsigned char *)strrcpy((char *)buf, line);
 			/* Put the new line back in */
 			return (unsigned char *)strrcpy((char *)buf, "\n");
 
 		case QUOTEDPRINTABLE:
+			if(line == NULL) {	/* empty line */
+				*buf++ = '\n';
+				break;
+			}
+			
 			softbreak = FALSE;
 			while(*line) {
 				if(*line == '=') {
@@ -1509,6 +1546,8 @@ decodeLine(const message *m, const char *line, unsigned char *buf, size_t buflen
 			break;
 
 		case BASE64:
+			if(line == NULL)
+				break;
 			/*
 			 * RFC1521 sets the maximum length to 76 bytes
 			 * but many e-mail clients ignore that
@@ -1523,14 +1562,14 @@ decodeLine(const message *m, const char *line, unsigned char *buf, size_t buflen
 			/*
 			 * Klez doesn't always put "=" on the last line
 			 */
-			/*buf = decode(line, buf, base64, p2 == NULL);*/
-			buf = decode(copy, buf, base64, FALSE);
+			buf = decode(copy, buf, base64, (p2 == NULL) && ((strlen(copy) & 3) == 0));
+			/*buf = decode(copy, buf, base64, FALSE);*/
 
 			free(copy);
 			break;
 
 		case UUENCODE:
-			if(*line == '\0')	/* empty line */
+			if((line == NULL) || (*line == '\0'))	/* empty line */
 				break;
 			if(strncasecmp(line, "begin ", 6) == 0)
 				break;
@@ -1548,7 +1587,7 @@ decodeLine(const message *m, const char *line, unsigned char *buf, size_t buflen
 				 * the maximum length of a uuencoded line is
 				 * 62 characters
 				 */
-				cli_warnmsg("uudecode: buffer overflow stopped, attempting to ignore but decoding may fail");
+				cli_warnmsg("uudecode: buffer overflow stopped, attempting to ignore but decoding may fail\n");
 			else
 				buf = decode(line, buf, uudecode, (len & 3) == 0);
 			break;
@@ -1566,6 +1605,7 @@ decode(const char *in, unsigned char *out, unsigned char (*decoder)(char), bool
 {
 	unsigned char b1, b2, b3, b4;
 
+	/*cli_dbgmsg("decode %s (len %d ifFast %d)\n", in, strlen(in), isFast);*/
 	if(isFast)
 		/* Fast decoding if not last line */
 		while(*in) {
@@ -1595,23 +1635,17 @@ decode(const char *in, unsigned char *out, unsigned char (*decoder)(char), bool
 				b2 = '\0';
 				nbytes = 1;
 			} else {
-				assert(*in != '\0');
-
 				b2 = (*decoder)(*in++);
 				if(*in == '\0') {
 					b3 = '\0';
 					nbytes = 2;
 				} else {
-					assert(*in != '\0');
-
 					b3 = (*decoder)(*in++);
 
 					if(*in == '\0') {
 						b4 = '\0';
 						nbytes = 3;
 					} else {
-						assert(*in != '\0');
-
 						b4 = (*decoder)(*in++);
 						nbytes = 4;
 					}
